#include "TSocket.h"
#include "TNtuple.h"
#include "TTree.h"
#include "TNetMaker.h"

TNetMaker::TNetMaker(const Char_t* host, Int_t port, Int_t buff_size) :
fBufferLength(0),
fBufferEmpty(true),
fBuffer(NULL),
fEndOfData(false)
{
	SetAddress(host, port);
	SetBufferLength(buff_size);
	SetLocalPort(20000);
}

TNetMaker::~TNetMaker()
{
	delete [] fBuffer;
}

TNetMaker::SetAddress(const Char_t* host, Int_t port)
{
	fHostName = host;
	fRemotePort = port;
}

Int_t TNetMaker::GetBufferLength()
{
	return (fBufferLength - 2 * sizeof(int)) / 1024;
}

void TNetMaker::SetBufferLength(Int_t length)
{
	if (GetBufferLength() != length)
    {
		if (!fBufferEmpty) Flush();
		if (fBuffer) delete [] fBuffer;

		fBufferEmpty = true;
		fBufferSpace = 1024 * length;
		fBufferLength = 2 * sizeof(Int_t) + fBufferSpace;
		fBuffer = new Char_t[fBufferLength];
		fBufferPos = 2 * sizeof(int);

		Int_t* stamp = (Int_t*)fBuffer;
		*stamp = 0xABCD;

		Int_t* counter = (Int_t*)(fBuffer+4);
		*counter = 0;
	}
}

Int_t TNetMaker::Send(Float_t* vector_f, Int_t length, Char_t type)
{
	Int_t header_length = SERIALIZED_HDR_SIZE;
	Int_t total_length = header_length + length*sizeof(Float_t);
	Char_t* serialized = new Char_t[total_length];
	serialized[0] = 0xE0;
	serialized[1] = type;

	Int_t* sizes = (Int_t*)(serialized+2);
	sizes[0] = length; sizes[1] = 0; sizes[2] = 0; sizes[3] = 0; sizes[4] = 0;

	Float_t* data = (Float_t*)(serialized+header_length);
	for (int i = 0; i < length; i++) data[i] = vector_f[i];

	if (fBufferSpace < total_length)
	{
		if (Flush() <= 0)
		{
			cout << "TNetMaker::Send(): cannot send data." << endl;
			delete [] serialized;
			return 0;
		}
	}
	if (!Queue(serialized, total_length))
	{
		cout << "TNetMaker::Send(): data buffer problem." << endl;
		delete [] serialized;
		return 0;
	}

	delete [] serialized;
	return 1;
}

Int_t TNetMaker::Send(Float_t* vector_f, Int_t inputs, Int_t targets,
		Int_t outputs, Int_t notused, Char_t type)
{
	Int_t header_length = SERIALIZED_HDR_SIZE;
	Int_t length = inputs + targets + outputs + notused;
	Int_t total_length = header_length + length*sizeof(Float_t);
	Char_t* serialized = new Char_t[total_length];
	serialized[0] = 0xE0;
	serialized[1] = type;

	if (outputs > 0) serialized[0] |= 1;

	Int_t* sizes = (Int_t*)(serialized+2);
	sizes[0] = inputs; sizes[1] = outputs; sizes[2] = notused; sizes[3] = targets; sizes[4] = 0;

	Float_t* data = (Float_t*)(serialized+header_length);
	for (int i = 0; i < length; i++) data[i] = vector_f[i];

	if (fBufferSpace < total_length)
	{
		if (Flush() <= 0)
		{
			cout << "TNetMaker::Send(): cannot send data." << endl;
			delete [] serialized;
			return 0;
		}
	}
	if (!Queue(serialized, total_length))
	{
		cout << "TNetMaker::Send(): data buffer problem." << endl;
		delete [] serialized;
		return 0;
	}

	delete [] serialized;
	return 1;
}

Int_t TNetMaker::SendTrigg()
{
	Int_t cmd_length = SERIALIZED_HDR_SIZE;
	Char_t* trigg_cmd = new Char_t[cmd_length];
    trigg_cmd[0] = 0x70;
	for (int i = 1; i < cmd_length; i++) trigg_cmd[i] = 0;

	if (fBufferSpace < cmd_length)
	{
		if (Flush() <= 0)
		{
			cout << "TNetMaker::SendTrigg(): cannot send data." << endl;
			delete [] trigg_cmd;
			return 0;
		}
	}
	if (!Queue(trigg_cmd, cmd_length))
	{
		cout << "TNetMaker::SendTrigg(): data buffer problem." << endl;
		delete [] trigg_cmd;
		return 0;
	}
	if (Flush() <= 0)
	{
		cout << "TNetMaker::SendTrigg(): cannot send command." << endl;
		delete [] trigg_cmd;
		return 0;
	}

	delete [] trigg_cmd;
	return 1;
}

Int_t TNetMaker::Send(TNtuple* ntuple, Char_t type)
{
	Int_t length = ntuple->GetNvar();
	Int_t nsamples = ntuple->GetEntries();
	for (Int_t i = 0; i < nsamples; i++)
	{
		ntuple->GetEntry(i);
		if (!Send(ntuple->GetArgs(), length, type))
		{
			cout << "TNetMaker::Send(): event #" << (i+1) << " not queued." << endl;
			return i;
		}
	}
	if (!SendTrigg()) cout << "TNetMaker::Send(): trigger command not sent." << endl;
	return nsamples;
}

Int_t TNetMaker::Send(TTree* tree, const Char_t* inputs,
		const Char_t* targets, const Char_t* outputs,
		const Char_t* notused, Char_t type)
{
	Int_t vi, index;
	Char_t var_name[255];

	Float_t fvar_buf[1024];
	Double_t dvar_buf[1024];
	Int_t ivar_buf[1024];
	Int_t buf_type[1024];
	Int_t buf_index = 0;

	Int_t ninputs = 0;
	vi = 0; index = 0;
	while (buf_index < 1024)
	{
		if (((inputs[index] == ':') || (inputs[index] == 0)) && (vi > 0))
		{
			var_name[vi] = 0;
			if (var_name[vi-2] == '/')
			{
				var_name[vi-2] = 0;
				switch (var_name[vi-1])
				{
				case 'f':
				case 'F':
					tree->SetBranchAddress(var_name, fvar_buf + buf_index);
					buf_type[buf_index] = 0;
					buf_index++; ninputs++;
					break;
				case 'd':
				case 'D':
					tree->SetBranchAddress(var_name, dvar_buf + buf_index);
					buf_type[buf_index] = 2;
					buf_index++; ninputs++;
					break;
				case 'i':
				case 'I':
					tree->SetBranchAddress(var_name, ivar_buf + buf_index);
					buf_type[buf_index] = 1;
					buf_index++; ninputs++;
					break;
				default:
					cout << "TNetMaker::Send(): variable type not supported." << endl;
					break;
				}
			}
			else
			{
				tree->SetBranchAddress(var_name, fvar_buf + buf_index);
				buf_type[buf_index] = 0;
				buf_index++; ninputs++;
			}
			vi = 0;
		}
		else
		{
			var_name[vi] = inputs[index];
			vi++;
		}

		if (inputs[index] == 0) break;
		index++;
	}

	Int_t noutputs = 0;
	if (outputs)
	{
		vi = 0; index = 0;
		while (buf_index < 1024)
		{
			if (((outputs[index] == ':') || (outputs[index] == 0)) && (vi > 0))
			{
				var_name[vi] = 0;
				if (var_name[vi-2] == '/')
				{
					var_name[vi-2] = 0;
					switch (var_name[vi-1])
					{
					case 'f':
					case 'F':
						tree->SetBranchAddress(var_name, fvar_buf + buf_index);
						buf_type[buf_index] = 0;
						buf_index++; noutputs++;
						break;
					case 'd':
					case 'D':
						tree->SetBranchAddress(var_name, dvar_buf + buf_index);
						buf_type[buf_index] = 2;
						buf_index++; noutputs++;
						break;
					case 'i':
					case 'I':
						tree->SetBranchAddress(var_name, ivar_buf + buf_index);
						buf_type[buf_index] = 1;
						buf_index++; noutputs++;
						break;
					default:
						cout << "TNetMaker::Send(): variable type not supported." << endl;
						break;
					}
				}
				else
				{
					tree->SetBranchAddress(var_name, fvar_buf + buf_index);
					buf_type[buf_index] = 0;
					buf_index++; noutputs++;
				}
				vi = 0;
			}
			else
			{
				var_name[vi] = outputs[index];
				vi++;
			}

			if (outputs[index] == 0) break;
			index++;
		}
	}

	Int_t nnotused = 0;
	if (notused)
	{
		vi = 0; index = 0;
		while (buf_index < 1024)
		{
			if (((notused[index] == ':') || (notused[index] == 0)) && (vi > 0))
			{
				var_name[vi] = 0;
				if (var_name[vi-2] == '/')
				{
					var_name[vi-2] = 0;
					switch (var_name[vi-1])
					{
					case 'f':
					case 'F':
						tree->SetBranchAddress(var_name, fvar_buf + buf_index);
						buf_type[buf_index] = 0;
						buf_index++; nnotused++;
						break;
					case 'd':
					case 'D':
						tree->SetBranchAddress(var_name, dvar_buf + buf_index);
						buf_type[buf_index] = 1;
						buf_index++; nnotused++;
						break;
					case 'i':
					case 'I':
						tree->SetBranchAddress(var_name, ivar_buf + buf_index);
						buf_type[buf_index] = 1;
						buf_index++; nnotused++;
						break;
					default:
						cout << "TNetMaker::Send(): variable type not supported." << endl;
						break;
					}
				}
				else
				{
					tree->SetBranchAddress(var_name, fvar_buf + buf_index);
					buf_type[buf_index] = 0;
					buf_index++; nnotused++;
				}
				vi = 0;
			}
			else
			{
				var_name[vi] = notused[index];
				vi++;
			}

			if (notused[index] == 0) break;
			index++;
		}
	}

	Int_t ntargets = 0;
	if (targets)
	{
		vi = 0; index = 0;
		while (buf_index < 1024)
		{
			if (((targets[index] == ':') || (targets[index] == 0)) && (vi > 0))
			{
				var_name[vi] = 0;
				if (var_name[vi-2] == '/')
				{
					var_name[vi-2] = 0;
					switch (var_name[vi-1])
					{
					case 'f':
					case 'F':
						tree->SetBranchAddress(var_name, fvar_buf + buf_index);
						buf_type[buf_index] = 0;
						buf_index++; ntargets++;
						break;
					case 'd':
					case 'D':
						tree->SetBranchAddress(var_name, dvar_buf + buf_index);
						buf_type[buf_index] = 1;
						buf_index++; ntargets++;
						break;
					case 'i':
					case 'I':
						tree->SetBranchAddress(var_name, ivar_buf + buf_index);
						buf_type[buf_index] = 1;
						buf_index++; ntargets++;
						break;
					default:
						cout << "TNetMaker::Send(): variable type not supported." << endl;
						break;
					}
				}
				else
				{
					tree->SetBranchAddress(var_name, fvar_buf + buf_index);
					buf_type[buf_index] = 0;
					buf_index++; ntargets++;
				}
				vi = 0;
			}
			else
			{
				var_name[vi] = targets[index];
				vi++;
			}

			if (targets[index] == 0) break;
			index++;
		}
	}

	Int_t length = ninputs + noutputs + nnotused + ntargets;
	Int_t nsamples = tree->GetEntries();
	for (Int_t i = 0; i < nsamples; i++)
	{
		tree->GetEntry(i);
		for (Int_t j = 0; j < length; j++)
		{
			if (buf_type[j] == 1)
			{
				fvar_buf[j] = (float)ivar_buf[j];
			}
			else if (buf_type[j] == 2)
			{
				fvar_buf[j] = (float)dvar_buf[j];
			}
		}
		if (!Send(fvar_buf, ninputs, ntargets, noutputs, nnotused, type))
		{
			cout << "TNetMaker::Send(): event #" << (i+1) << " not queued." << endl;
			return i;
		}
	}
	if (!SendTrigg()) cout << "TNetMaker::Send(): trigger command not sent." << endl;
	return nsamples;
}

Int_t TNetMaker::Queue(Char_t* bytes, Int_t length)
{
	if (length > fBufferSpace)
	{
		cout << "TNetMaker::Queue() buffer overrun." << endl;
		return 0;
	}

	for (int i = 0; i < length; i++, fBufferPos++)
		fBuffer[fBufferPos] = bytes[i];

	Int_t* counter = (Int_t*)(fBuffer+4);
	(*counter)++;

	fBufferSpace -= length;
	fBufferEmpty = false;
	return 1;
}

Int_t TNetMaker::Flush()
{
	if (fBufferEmpty) return 0;
	if (!fHostName)
	{
		cout << "TNetMaker::Flush() host is not set." << endl;
		return -1;
	}
	if (!fRemotePort)
	{
		cout << "TNetMaker::Flush() port is not set." << endl;
		return -1;
	}

	TSocket* socket = new TSocket(fHostName, fRemotePort);
	if (!socket || !socket->IsValid())
	{
		cout << "TNetMaker::Flush() socket problem." << endl;
		return -1;
	}

	Char_t str[32];
	socket->RecvRaw(str, 4);
    str[0] = (Char_t)(fBufferPos & 0xFF);
    str[1] = (Char_t)((fBufferPos >> 8) & 0xFF);
    str[2] = (Char_t)((fBufferPos >> 16) & 0xFF);
    str[3] = (Char_t)((fBufferPos >> 24) & 0xFF);
	socket->SendRaw(str, 4);
	socket->RecvRaw(str, 4);

	int b = socket->SendRaw(fBuffer, fBufferPos);
	if (b < 0)
		cout << "TNetMaker::Flush() sending error code " << b << "." << endl;
	else if (b != fBufferPos)
		cout << "Only " << b << " bytes sent out of " << fBufferPos << " bytes." << endl;

	socket->RecvRaw(str, 4);
	socket->Close();

	fBufferSpace = fBufferLength - 2 * sizeof(int);
    fBufferPos = 2 * sizeof(int);
    fBufferEmpty = true;

	Int_t* counter = (Int_t*)(fBuffer+4);
	*counter = 0;
	return b;
}

TNtuple* TNetMaker::GetNtuple()
{
	TServerSocket* ss = new TServerSocket(fLocalPort, kTRUE);
	TSocket* socket;
	fEndOfData = false;
	TNtuple* ntuple = NULL;
	Char_t* str = new Char_t[16];
	Char_t* buffer;
	Int_t k, b, len, rcvLength;

	while (!fEndOfData)
	{
		socket = ss->Accept();

		sprintf(str, "go!");
		socket->SendRaw(str, 4);
		socket->RecvRaw(str, 4);
		rcvLength = *((Int_t*)str);
		buffer = new Char_t[rcvLength];
		sprintf(str, "dta"); str[3] = 0;
		socket->SendRaw(str, 4);

		b = 0;
		len = rcvLength;
		do
		{
			k = socket->RecvRaw(buffer + b, len);
			b += k; len -= k;
		}
		while (b < rcvLength);

		sprintf(str, "ok."); str[3] = 0;
		socket->SendRaw(str, 4);
		gSystem->Sleep(50);

		k = ParseData(buffer, &ntuple);

		delete [] buffer;
		socket->Close();
		delete socket;
	}
	ss->Close();
	delete ss;
	delete [] str;
	return ntuple;
}

Int_t TNetMaker::ParseData(Char_t* buffer, TNtuple** ntuple)
{
	if (*((Int_t*)buffer) != 0xABCD)
	{
		cout << "TNetMaker::ParseData(): ID does not match." << endl;
		return 0;
	}
	Int_t nsamples = *((Int_t*)(buffer + sizeof(Int_t)));

	TNtuple* nt = *ntuple;
	Int_t headerLength = SERIALIZED_HDR_SIZE;
    Int_t vLength, bufPos = 2 * sizeof(Int_t);
    for (Int_t s = 0; s < nsamples; s++)
    {
		UChar_t sType = (UChar_t)(buffer[bufPos] & 0xF0);
        switch (sType)
        {
		case 0xE0: // data sample
			Int_t* intPt = (Int_t*)(buffer + bufPos + 2);
			if (nt == NULL)
			{
				vLength = intPt[0] + intPt[1] + intPt[2] + intPt[3] + intPt[4];
				char* varlist = new char[vLength * 8];
				memset(varlist, 0, vLength * 8);
				sprintf(varlist, "i1");
				for (int i = 1; i < intPt[0]; i++)
					sprintf(varlist + strlen(varlist), ":i%d", i+1);
				for (int o = 0; o < intPt[1]; o++)
					sprintf(varlist + strlen(varlist), ":o%d", o+1);
				for (int n = 0; n < intPt[2]; n++)
					sprintf(varlist + strlen(varlist), ":n%d", n+1);
				for (int t = 0; t < intPt[3]; t++)
					sprintf(varlist + strlen(varlist), ":t%d", t+1);
				for (int b = 0; b < intPt[4]; b++)
					sprintf(varlist + strlen(varlist), ":b%d", t+1);
				nt = new TNtuple("NetMaker", "Processed data", varlist);
				delete [] varlist;
				if (nt->GetNvar() != vLength)
				{
					cout << "TNetMaker::ParseData(): ntuple problem." << endl;
					return 0;
				}
			}
			else
			{
				vLength = nt->GetNvar();
				if (vLength != intPt[0] + intPt[1] + intPt[2] + intPt[3] + intPt[4])
				{
					cout << "TNetMaker::ParseData(): vector lenght does not match." << endl;
					return s;
				}
			}
			nt->Fill((Float_t*)(buffer + bufPos + headerLength));
			Int_t dataLength = vLength * sizeof(Float_t);
            bufPos += headerLength + dataLength;
            break;
		case 0x70: // trigg command
			bufPos += headerLength;
			fEndOfData = true;
			break;
		default:
			cout << "TNetMaker::ParseData(): data format not recognized." << endl;
			return s;
		}
		if (fEndOfData) break;
	}

	*ntuple = nt;
	return nsamples;
}
